@@ -55,7 +55,7 @@ static void Help(void) {
   printf("Usage:\n");
   printf(" gif2webp [options] gif_file -o webp_file\n");
   printf("Options:\n");
-  printf("  -h / -help  ............ this help\n");
+  printf("  -h / -help ............. this help\n");
   printf("  -lossy ................. encode image using lossy compression\n");
   printf("  -mixed ................. for each frame in the image, pick lossy\n"
          "                           or lossless compression heuristically\n");
@@ -98,7 +98,6 @@ int main(int argc, const char *argv[]) {
   WebPPicture frame;                // Frame rectangle only (not disposed).
   WebPPicture curr_canvas;          // Not disposed.
   WebPPicture prev_canvas;          // Disposed.
-  WebPPicture prev_to_prev_canvas;  // Disposed.
 
   WebPAnimEncoder* enc = NULL;
   WebPAnimEncoderOptions enc_options;
@@ -124,8 +123,7 @@ int main(int argc, const char *argv[]) {
 
   if (!WebPConfigInit(&config) || !WebPAnimEncoderOptionsInit(&enc_options) ||
       !WebPPictureInit(&frame) || !WebPPictureInit(&curr_canvas) ||
-      !WebPPictureInit(&prev_canvas) ||
-      !WebPPictureInit(&prev_to_prev_canvas)) {
+      !WebPPictureInit(&prev_canvas)) {
     fprintf(stderr, "Error! Version mismatch!\n");
     return -1;
   }
@@ -217,6 +215,7 @@ int main(int argc, const char *argv[]) {
       return 0;
     } else if (!strcmp(argv[c], "-quiet")) {
       quiet = 1;
+      enc_options.verbose = 0;
     } else if (!strcmp(argv[c], "-v")) {
       verbose = 1;
       enc_options.verbose = 1;
@@ -304,7 +303,6 @@ int main(int argc, const char *argv[]) {
           GIFClearPic(&frame, NULL);
           WebPPictureCopy(&frame, &curr_canvas);
           WebPPictureCopy(&frame, &prev_canvas);
-          WebPPictureCopy(&frame, &prev_to_prev_canvas);
 
           // Background color.
           GIFGetBackgroundColor(gif->SColorMap, gif->SBackGroundColor,
@@ -341,7 +339,6 @@ int main(int argc, const char *argv[]) {
         }
 
         // Update canvases.
-        GIFCopyPixels(&prev_canvas, &prev_to_prev_canvas);
         GIFDisposeFrame(orig_dispose, &gif_rect, &prev_canvas, &curr_canvas);
         GIFCopyPixels(&curr_canvas, &prev_canvas);
 
@@ -362,6 +359,8 @@ int main(int argc, const char *argv[]) {
         if (DGifGetExtension(gif, &extension, &data) == GIF_ERROR) {
           goto End;
         }
+        if (data == NULL) continue;
+
         switch (extension) {
           case COMMENT_EXT_FUNC_CODE: {
             break;  // Do nothing for now.
@@ -529,7 +528,6 @@ int main(int argc, const char *argv[]) {
   WebPPictureFree(&frame);
   WebPPictureFree(&curr_canvas);
   WebPPictureFree(&prev_canvas);
-  WebPPictureFree(&prev_to_prev_canvas);
   WebPAnimEncoderDelete(enc);
   if (out != NULL && out_file != NULL) fclose(out);
 
